@@ -26,12 +26,12 @@
  * are passed to the free space map at conclusion of the scan.
  *
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/commands/vacuumlazy.c,v 1.11 2002/01/06 00:37:44 tgl Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/commands/vacuumlazy.c,v 1.20.2.1 2005/05/07 21:33:47 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -45,6 +45,7 @@
 #include "storage/freespace.h"
 #include "storage/sinval.h"
 #include "storage/smgr.h"
+#include "utils/lsyscache.h"
 
 
 /*
@@ -86,13 +87,12 @@ typedef struct LVRelStats
 	/* We use a simple array until it fills up, then convert to heap */
 	bool		fs_is_heap;		/* are we using heap organization? */
 	int			num_free_pages; /* current # of entries */
-	int			max_free_pages; /* # slots allocated in arrays */
-	BlockNumber *free_pages;	/* array or heap of block numbers */
-	Size	   *free_spaceavail;	/* array or heap of available space */
+	int			max_free_pages; /* # slots allocated in array */
+	PageFreeSpaceInfo *free_pages;	/* array or heap of blkno/avail */
 } LVRelStats;
 
 
-static int	MESSAGE_LEVEL;		/* message level */
+static int	elevel = -1;
 
 static TransactionId OldestXmin;
 static TransactionId FreezeLimit;
@@ -118,6 +118,7 @@ static bool lazy_tid_reaped(ItemPointer itemptr, void *state);
 static bool dummy_tid_reaped(ItemPointer itemptr, void *state);
 static void lazy_update_fsm(Relation onerel, LVRelStats *vacrelstats);
 static int	vac_cmp_itemptr(const void *left, const void *right);
+static int	vac_cmp_page_spaces(const void *left, const void *right);
 
 
 /*
@@ -138,11 +139,10 @@ lazy_vacuum_rel(Relation onerel, VacuumStmt *vacstmt)
 	bool		hasindex;
 	BlockNumber possibly_freeable;
 
-	/* initialize */
 	if (vacstmt->verbose)
-		MESSAGE_LEVEL = NOTICE;
+		elevel = INFO;
 	else
-		MESSAGE_LEVEL = DEBUG;
+		elevel = DEBUG1;
 
 	vacuum_set_xid_limits(vacstmt, onerel->rd_rel->relisshared,
 						  &OldestXmin, &FreezeLimit);
@@ -208,7 +208,9 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 	vac_init_rusage(&ru0);
 
 	relname = RelationGetRelationName(onerel);
-	elog(MESSAGE_LEVEL, "--Relation %s--", relname);
+	elog(elevel, "--Relation %s.%s--",
+		 get_namespace_name(RelationGetNamespace(onerel)),
+		 relname);
 
 	empty_pages = changed_pages = 0;
 	num_tuples = tups_vacuumed = nkeep = nunused = 0;
@@ -259,12 +261,34 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 
 		if (PageIsNew(page))
 		{
-			/* Not sure we still need to handle this case, but... */
+			/*
+			 * An all-zeroes page could be left over if a backend extends
+			 * the relation but crashes before initializing the page.
+			 * Reclaim such pages for use.
+			 *
+			 * We have to be careful here because we could be looking at
+			 * a page that someone has just added to the relation and not
+			 * yet been able to initialize (see RelationGetBufferForTuple).
+			 * To interlock against that, release the buffer read lock
+			 * (which we must do anyway) and grab the relation extension
+			 * lock before re-locking in exclusive mode.  If the page is
+			 * still uninitialized by then, it must be left over from a
+			 * crashed backend, and we can initialize it.
+			 *
+			 * We don't really need the relation lock when this is a new
+			 * or temp relation, but it's probably not worth the code space
+			 * to check that, since this surely isn't a critical path.
+			 *
+			 * Note: the comparable code in vacuum.c need not do all this
+			 * because it's got exclusive lock on the whole relation.
+			 */
 			LockBuffer(buf, BUFFER_LOCK_UNLOCK);
+			LockPage(onerel, 0, ExclusiveLock);
+			UnlockPage(onerel, 0, ExclusiveLock);
 			LockBuffer(buf, BUFFER_LOCK_EXCLUSIVE);
 			if (PageIsNew(page))
 			{
-				elog(NOTICE, "Rel %s: Uninitialized page %u - fixing",
+				elog(WARNING, "Rel %s: Uninitialized page %u - fixing",
 					 relname, blkno);
 				PageInit(page, BufferGetPageSize(buf), 0);
 				lazy_record_free_space(vacrelstats, blkno,
@@ -330,11 +354,11 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 					 * assumption by momentarily acquiring exclusive lock,
 					 * but for the moment I see no need to.
 					 */
-					if (TransactionIdIsNormal(tuple.t_data->t_xmin) &&
-						TransactionIdPrecedes(tuple.t_data->t_xmin,
+					if (TransactionIdIsNormal(HeapTupleHeaderGetXmin(tuple.t_data)) &&
+						TransactionIdPrecedes(HeapTupleHeaderGetXmin(tuple.t_data),
 											  FreezeLimit))
 					{
-						tuple.t_data->t_xmin = FrozenTransactionId;
+						HeapTupleHeaderSetXmin(tuple.t_data, FrozenTransactionId);
 						/* infomask should be okay already */
 						Assert(tuple.t_data->t_infomask & HEAP_XMIN_COMMITTED);
 						pgchanged = true;
@@ -366,9 +390,9 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 			/*
 			 * Other checks...
 			 */
-			if (!OidIsValid(tuple.t_data->t_oid) &&
-				onerel->rd_rel->relhasoids)
-				elog(NOTICE, "Rel %s: TID %u/%u: OID IS INVALID. TUPGONE %d.",
+			if (onerel->rd_rel->relhasoids &&
+				!OidIsValid(HeapTupleGetOid(&tuple)))
+				elog(WARNING, "Rel %s: TID %u/%u: OID IS INVALID. TUPGONE %d.",
 					 relname, blkno, offnum, (int) tupgone);
 
 			if (tupgone)
@@ -430,8 +454,7 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 			lazy_scan_index(Irel[i], vacrelstats);
 	}
 
-	elog(MESSAGE_LEVEL, "Pages %u: Changed %u, Empty %u; \
-Tup %.0f: Vac %.0f, Keep %.0f, UnUsed %.0f.\n\tTotal %s",
+	elog(elevel, "Pages %u: Changed %u, Empty %u; Tup %.0f: Vac %.0f, Keep %.0f, UnUsed %.0f.\n\tTotal %s",
 		 nblocks, changed_pages, empty_pages,
 		 num_tuples, tups_vacuumed, nkeep, nunused,
 		 vac_show_rusage(&ru0));
@@ -481,8 +504,7 @@ lazy_vacuum_heap(Relation onerel, LVRelStats *vacrelstats)
 		npages++;
 	}
 
-	elog(MESSAGE_LEVEL, "Removed %d tuples in %d pages.\n\t%s",
-		 tupindex, npages,
+	elog(elevel, "Removed %d tuples in %d pages.\n\t%s", tupindex, npages,
 		 vac_show_rusage(&ru0));
 }
 
@@ -522,6 +544,8 @@ lazy_vacuum_page(Relation onerel, BlockNumber blkno, Buffer buffer,
 
 	uncnt = PageRepairFragmentation(page, unused);
 
+	/* XLOG stuff */
+	if (!onerel->rd_istemp)
 	{
 		XLogRecPtr	recptr;
 
@@ -530,6 +554,12 @@ lazy_vacuum_page(Relation onerel, BlockNumber blkno, Buffer buffer,
 		PageSetLSN(page, recptr);
 		PageSetSUI(page, ThisStartUpID);
 	}
+	else
+	{
+		/* No XLOG record, but still need to flag that XID exists on disk */
+		MyXactMadeTempRelUpdate = true;
+	}
+
 	END_CRIT_SECTION();
 
 	return tupindex;
@@ -589,7 +619,7 @@ lazy_scan_index(Relation indrel, LVRelStats *vacrelstats)
 						stats->num_pages, stats->num_index_tuples,
 						false);
 
-	elog(MESSAGE_LEVEL, "Index %s: Pages %u; Tuples %.0f.\n\t%s",
+	elog(elevel, "Index %s: Pages %u; Tuples %.0f.\n\t%s",
 		 RelationGetRelationName(indrel),
 		 stats->num_pages, stats->num_index_tuples,
 		 vac_show_rusage(&ru0));
@@ -636,7 +666,7 @@ lazy_vacuum_index(Relation indrel, LVRelStats *vacrelstats)
 							stats->num_pages, stats->num_index_tuples,
 							false);
 
-		elog(MESSAGE_LEVEL, "Index %s: Pages %u; Tuples %.0f: Deleted %.0f.\n\t%s",
+		elog(elevel, "Index %s: Pages %u; Tuples %.0f: Deleted %.0f.\n\t%s",
 			 RelationGetRelationName(indrel), stats->num_pages,
 			 stats->num_index_tuples, stats->tuples_removed,
 			 vac_show_rusage(&ru0));
@@ -653,8 +683,7 @@ lazy_truncate_heap(Relation onerel, LVRelStats *vacrelstats)
 {
 	BlockNumber old_rel_pages = vacrelstats->rel_pages;
 	BlockNumber new_rel_pages;
-	BlockNumber *pages;
-	Size	   *spaceavail;
+	PageFreeSpaceInfo *pageSpaces;
 	int			n;
 	int			i,
 				j;
@@ -727,28 +756,27 @@ lazy_truncate_heap(Relation onerel, LVRelStats *vacrelstats)
 	 * Drop free-space info for removed blocks; these must not get entered
 	 * into the FSM!
 	 */
-	pages = vacrelstats->free_pages;
-	spaceavail = vacrelstats->free_spaceavail;
+	pageSpaces = vacrelstats->free_pages;
 	n = vacrelstats->num_free_pages;
 	j = 0;
 	for (i = 0; i < n; i++)
 	{
-		if (pages[i] < new_rel_pages)
+		if (pageSpaces[i].blkno < new_rel_pages)
 		{
-			pages[j] = pages[i];
-			spaceavail[j] = spaceavail[i];
+			pageSpaces[j] = pageSpaces[i];
 			j++;
 		}
 	}
 	vacrelstats->num_free_pages = j;
+	/* We destroyed the heap ordering, so mark array unordered */
+	vacrelstats->fs_is_heap = false;
 
 	/*
 	 * We keep the exclusive lock until commit (perhaps not necessary)?
 	 */
 
-	elog(MESSAGE_LEVEL, "Truncated %u --> %u pages.\n\t%s",
-		 old_rel_pages, new_rel_pages,
-		 vac_show_rusage(&ru0));
+	elog(elevel, "Truncated %u --> %u pages.\n\t%s", old_rel_pages,
+		 new_rel_pages, vac_show_rusage(&ru0));
 }
 
 /*
@@ -905,10 +933,8 @@ lazy_space_alloc(LVRelStats *vacrelstats, BlockNumber relblocks)
 	vacrelstats->fs_is_heap = false;
 	vacrelstats->num_free_pages = 0;
 	vacrelstats->max_free_pages = maxpages;
-	vacrelstats->free_pages = (BlockNumber *)
-		palloc(maxpages * sizeof(BlockNumber));
-	vacrelstats->free_spaceavail = (Size *)
-		palloc(maxpages * sizeof(Size));
+	vacrelstats->free_pages = (PageFreeSpaceInfo *)
+		palloc(maxpages * sizeof(PageFreeSpaceInfo));
 }
 
 /*
@@ -938,8 +964,7 @@ lazy_record_free_space(LVRelStats *vacrelstats,
 					   BlockNumber page,
 					   Size avail)
 {
-	BlockNumber *pages;
-	Size	   *spaceavail;
+	PageFreeSpaceInfo *pageSpaces;
 	int			n;
 
 	/* Ignore pages with little free space */
@@ -947,15 +972,14 @@ lazy_record_free_space(LVRelStats *vacrelstats,
 		return;
 
 	/* Copy pointers to local variables for notational simplicity */
-	pages = vacrelstats->free_pages;
-	spaceavail = vacrelstats->free_spaceavail;
+	pageSpaces = vacrelstats->free_pages;
 	n = vacrelstats->max_free_pages;
 
 	/* If we haven't filled the array yet, just keep adding entries */
 	if (vacrelstats->num_free_pages < n)
 	{
-		pages[vacrelstats->num_free_pages] = page;
-		spaceavail[vacrelstats->num_free_pages] = avail;
+		pageSpaces[vacrelstats->num_free_pages].blkno = page;
+		pageSpaces[vacrelstats->num_free_pages].avail = avail;
 		vacrelstats->num_free_pages++;
 		return;
 	}
@@ -963,7 +987,7 @@ lazy_record_free_space(LVRelStats *vacrelstats,
 	/*----------
 	 * The rest of this routine works with "heap" organization of the
 	 * free space arrays, wherein we maintain the heap property
-	 *			spaceavail[(j-1) div 2] <= spaceavail[j]  for 0 < j < n.
+	 *			avail[(j-1) div 2] <= avail[j]  for 0 < j < n.
 	 * In particular, the zero'th element always has the smallest available
 	 * space and can be discarded to make room for a new page with more space.
 	 * See Knuth's discussion of heap-based priority queues, sec 5.2.3;
@@ -983,8 +1007,8 @@ lazy_record_free_space(LVRelStats *vacrelstats,
 
 		while (--l >= 0)
 		{
-			BlockNumber R = pages[l];
-			Size		K = spaceavail[l];
+			BlockNumber R = pageSpaces[l].blkno;
+			Size		K = pageSpaces[l].avail;
 			int			i;		/* i is where the "hole" is */
 
 			i = l;
@@ -994,23 +1018,22 @@ lazy_record_free_space(LVRelStats *vacrelstats,
 
 				if (j >= n)
 					break;
-				if (j + 1 < n && spaceavail[j] > spaceavail[j + 1])
+				if (j + 1 < n && pageSpaces[j].avail > pageSpaces[j + 1].avail)
 					j++;
-				if (K <= spaceavail[j])
+				if (K <= pageSpaces[j].avail)
 					break;
-				pages[i] = pages[j];
-				spaceavail[i] = spaceavail[j];
+				pageSpaces[i] = pageSpaces[j];
 				i = j;
 			}
-			pages[i] = R;
-			spaceavail[i] = K;
+			pageSpaces[i].blkno = R;
+			pageSpaces[i].avail = K;
 		}
 
 		vacrelstats->fs_is_heap = true;
 	}
 
 	/* If new page has more than zero'th entry, insert it into heap */
-	if (avail > spaceavail[0])
+	if (avail > pageSpaces[0].avail)
 	{
 		/*
 		 * Notionally, we replace the zero'th entry with the new data, and
@@ -1026,16 +1049,15 @@ lazy_record_free_space(LVRelStats *vacrelstats,
 
 			if (j >= n)
 				break;
-			if (j + 1 < n && spaceavail[j] > spaceavail[j + 1])
+			if (j + 1 < n && pageSpaces[j].avail > pageSpaces[j + 1].avail)
 				j++;
-			if (avail <= spaceavail[j])
+			if (avail <= pageSpaces[j].avail)
 				break;
-			pages[i] = pages[j];
-			spaceavail[i] = spaceavail[j];
+			pageSpaces[i] = pageSpaces[j];
 			i = j;
 		}
-		pages[i] = page;
-		spaceavail[i] = avail;
+		pageSpaces[i].blkno = page;
+		pageSpaces[i].avail = avail;
 	}
 }
 
@@ -1077,16 +1099,17 @@ dummy_tid_reaped(ItemPointer itemptr, void *state)
 static void
 lazy_update_fsm(Relation onerel, LVRelStats *vacrelstats)
 {
+	PageFreeSpaceInfo *pageSpaces = vacrelstats->free_pages;
+	int			nPages = vacrelstats->num_free_pages;
+
 	/*
-	 * Since MultiRecordFreeSpace doesn't currently impose any
-	 * restrictions on the ordering of the input, we can just pass it the
-	 * arrays as-is, whether they are in heap or linear order.
+	 * Sort data into order, as required by MultiRecordFreeSpace.
 	 */
-	MultiRecordFreeSpace(&onerel->rd_node,
-						 0, MaxBlockNumber,
-						 vacrelstats->num_free_pages,
-						 vacrelstats->free_pages,
-						 vacrelstats->free_spaceavail);
+	if (nPages > 1)
+		qsort(pageSpaces, nPages, sizeof(PageFreeSpaceInfo),
+			  vac_cmp_page_spaces);
+
+	MultiRecordFreeSpace(&onerel->rd_node, 0, nPages, pageSpaces);
 }
 
 /*
@@ -1118,3 +1141,16 @@ vac_cmp_itemptr(const void *left, const void *right)
 
 	return 0;
 }
+
+static int
+vac_cmp_page_spaces(const void *left, const void *right)
+{
+	PageFreeSpaceInfo *linfo = (PageFreeSpaceInfo *) left;
+	PageFreeSpaceInfo *rinfo = (PageFreeSpaceInfo *) right;
+
+	if (linfo->blkno < rinfo->blkno)
+		return -1;
+	else if (linfo->blkno > rinfo->blkno)
+		return 1;
+	return 0;
+}
